#!/usr/local/bin/perl5

# HTTP router/ad filter/cookie filter/cache shim daemon
# Copyright 1999 Phil Karn, KA9Q
# May be used under the terms of the GNU Public License

# Configuration parameters

$verbose = 1;	# Tracing level

#Routing/configuration file
$routingtable = "/home/karn/lib/httproute/routes";

#Path prefix to local cache; comment out to disable local caching
$cachepath = "/crypt/web";

# Local port on which we'll listen (set up Netscape's proxy to point
# to 127.0.0.1 and this port)
$localport = 8080;

# Read buffer size for data relays
$bufsize = 512;

#end of local configuration parameters

use FileHandle;
use Socket;
use Time::Local;

$proto = getprotobyname('tcp');

$SIG{CHLD} = sub { wait };
$SIG{HUP} = \&rtsetup;


%monum =
(
 "Jan" => 0,
 "Feb" => 1,
 "Mar" => 2,
 "Apr" => 3,
 "May" => 4,
 "Jun" => 5,
 "Jul" => 6,
 "Aug" => 7,
 "Sep" => 8,
 "Oct" => 9,
 "Nov" => 10,
 "Dec" => 11,
);

rtsetup();

socket(Server, PF_INET, SOCK_STREAM, $proto) or die "socket";
setsockopt(Server, SOL_SOCKET, SO_REUSEADDR, pack("l", 1)) or warn "setsockopt";
bind(Server, sockaddr_in($localport,INADDR_ANY)) or die "bind";
listen(Server, SOMAXCONN) or die "listen";

# Accept requests, fork off servers to handle
for(;;){
  my $paddr,$pid;

  $paddr = accept(Client,Server);
  # strange double-fork kludge that avoids zombies, which the
  # $SIG{CHLD} reaper doesn't seem to do properly
  # see ftp://uiarchive.uiuc.edu/pub/lang/perl/CPAN/doc/manual/html/pod/perlfunc/fork.html
  unless($pid = fork){
    unless (fork) {
      doit();	# child
      exit 0;
    }
    exit 0;
  }
  waitpid($pid,0);
}	# parent never exits

# Handle requests in a single TCP connection
sub doit()
{
  my (@options,$url,$length,$cmd,$url,$version,$cookiesok);
  my $connection;
  
 proc:	while($_ = <Client>){
    # parse GET/POST/HEAD URL HTTP/version
    print STDERR $_ if($verbose > 1);
    if(!/^(\S+)\s+(\S+)\s+(\S+)/){
      print Client "HTTP/1.0 400 Bad Request\r\n";
      print STDERR "bad request\n" if($verbose > 1);
      goto done;
    }
    $cmd = $1;
    $url = $2;
    $version = $3;
    # accumulate option lines
    undef @options;
    undef $nocache;
    while($_ =  <Client>){
      push @options, $_;
      print STDERR $_  if($verbose > 2);
      if(/^Content-Length: ([0-9]+)/io){
	$length = $1;
      } elsif(/^Pragma:\s+no-cache/io){
	$nocache = 1;
      }
      goto scan if(/^\s+$/);
    }
    goto done;	# EOF from client, quit
    
  scan:;		# Scan the routing table
    # lower-case the method and host name
    $url =~ /([^:]+):\/\/([^\/]+)(.*)/;
    $urlprefix = lc($1) . '://' . lc($2) . $3;
    while(! $routetab { $urlprefix }){
      if($urlprefix =~ m$(\S+)/.*$){
	$urlprefix = $1;
      } else {
	last;	# no more slashes
      }
    }
    @entry = @{ $routetab { $urlprefix}};
    if(@entry){
      # Found a match
      $proxy = $entry[0];
      $cookiesok = $entry[1];
    } else {
      # default route
      $proxy = 1;
      $cookiesok = 0;
    }
    print STDERR "proxy $proxy cookiesok $cookiesok\n" if($verbose > 1);
    # Handle blacklisted ads
    undef $file;
    if($proxy == 0 and $adresp){
      $file = $adresp;
      print STDERR "Blanked       $cmd $url\n" if ($verbose > 0);
    } elsif(-d $cachepath){
      # lower case the method and host name, just in case
      $url =~ /([^:]+):\/\/([^\/]+)(.*)/;
      $file = $cachepath . '/' . lc($1) . '://' . lc($2) . $3;
      $file .= 'index.html' if($file =~ /\/$/);
    }
    if($proxy == 0 or (-f $file and ($cmd =~ m/get/i) and (not $nocache))){
      print STDERR "Local Hit $cmd $url\n" if ($verbose > 0 and $proxy != 0);
      open File, $file;
      undef $has_length;
      # pass header to client, looking for length
      do {
	$_ = <File>;
	$has_length = 1 if(/^Content-Length:/gio);
	print Client $_ or exit;
      } while(not /^$/);
      do {
	$piece = read File, $buf, $bufsize;
	print Client $buf or exit;
      } while($piece == $bufsize);
      undef $buf;
      close File;
      if($has_length){
	FileHandle::flush(Client);
	next proc;
      } else {
	# no content-length, must close connection
	goto done;
      }
    }
    print STDERR "Cache Miss\n" if ($verbose > 1);
    
    if($proxy ne $connection){
      close Proxy;
      socket(Proxy,PF_INET,SOCK_STREAM,$proto) or warn "socket: $!";
      connect(Proxy,$proxies[$proxy]) or warn "connect: $!";
      $connection = $proxy;
    }
    # relay (edited) HTTP options to proxy server
    print Proxy "$cmd $url $version\r\n" or exit;
    $length = 0;
    foreach (@options){
      $length = $1 if(/^Content-Length:\s+([0-9]+)/io);
      if(/^User-Agent:/io){
	# Lie that we're running something common
	print Proxy "User-Agent: Mozilla/3.02Gold (Win95; U)\r\n" or exit;
      } elsif(/^Referer:/io){
	# Lie about this one for privacy reasons
	$url =~ m/(http|ftp|gopher|wais):\/\/([^\/]*)(\/.*\/)?/io;
	$baseurl = "$1://$2";
	print Proxy "Referer: $baseurl\r\n" or exit;
      } else {
	print Proxy $_ or exit;
      }
    }
    undef @options;
    # relay HTTP body (if any, generally only on POST commands)
    while($length > 0){
      $chunk = min($length,$bufsize);
      read Client, $buf, $chunk;
      print Proxy $buf or exit;
      undef $buf;
      $length -= $chunk;
    }
    FileHandle::flush(Proxy);
    
    # Read response from proxy server
    $length = 0;
    undef @object;
    $timestamp = time;
    $_ = <Proxy>;
    push @object, $_;
    /HTTP\/([0-9\.]*)\s+([0-9]*)\s+(.*)/;
    $respcode = $2;
    print Client $_ or exit;
    print STDERR $_ if ($verbose > 2);
    $proxy_hit = '         ' if(not $nocache);
    $proxy_hit = 'No cache ' if $nocache;
    while($_ = <Proxy>){
      # Pass cookies only if OK
      print STDERR $_ if ($verbose > 2);
      if($cookiesok || !/^Set-Cookie: /io){
	print Client $_ or exit;
	push @object, $_;
      }
      $proxy_hit = "Proxy Hit" if((not $nocache) and /^X-Cache:\s+HIT/);
      $length = $1 if(/^Content-Length: ([0-9]+)/io);
      if(/^Last-Modified:\s+(.*)/i){
	$timestamp = parsedate($1);
      }
      last if(/^\s*$/);
    }
    if(not $_){
      # unexpected EOF from proxy
      close Client;
      undef $connection;
      goto done;
    }
    print STDERR "$proxy_hit $respcode $cmd $url\n" if ($verbose > 0);
    $count = 0;
    $left = $length;
    $left = 2e9 if ($left == 0);
    for(;$left != 0;$left -= $chunk){
      $chunk = min($left,$bufsize);
      last if not $cnt = read Proxy, $buf, $chunk;
      $count += $cnt;
      print Client $buf or exit;
      push @object, $buf;
    }
    FileHandle::flush(Client);
    if($file and $cmd =~ /get/i and $respcode == 200 and ($length == 0 or $length == $count)){
      # OK to save to cache file
      # Create any needed intermediate directories
      @dirs = split('/',$file);
      $partial = "";
      while(@dirs > 1){
	$partial .= (shift @dirs) . '/';
	if(-f $partial){
	  # rename as ordinary file within new dir
	  rename $partial, "$partial.tmp$$";
	  print STDERR "mkdir $partial\n";
	  mkdir $partial, 0777;
	  rename "$partial.tmp$$" , "$partial/index.html";
	} elsif(not -d $partial){
	  print STDERR "mkdir $partial\n";
	  mkdir $partial, 0777;
	}
      }
      undef @dirs;
      open File, ">$file$$";
      print File @object or exit;
      close File;
      utime $timestamp, $timestamp, "$file$$";
      # create it in an atomic operation
      rename "$file$$","$file";
    }
    undef @object;
    goto done if($length == 0);
  }
 done:	close Client;
  close Proxy if($connection);
}

sub min(){
  $x = shift;
  $y = shift;
  return ($x < $y) ? $x : $y;
}


sub rtsetup(){
  undef %files;
  undef @commands;
  readfile($routingtable);
  
  %routetab = ();
  @proxies= ();
  while($_ = shift @commands){
    chomp;
    if(/^$/){
      # blank line - skip
    } elsif(/^\#/){
      # comment - skip
    } elsif(/^defaultproxy\s+([0-9\.]+):([0-9]+)/io){
      $proxies[1] = sockaddr_in($2,inet_aton($1));
    } elsif(/^local\s+(.*)$/io){
      $adresp = $1;
      while($_ = shift @commands){
	chomp;
	if(/^$/io){
	  last;
	}
	$routetab{$_} = [0, 0];
      }
    } elsif(/^proxy\s+(\S+)/io){
      my ($proxy,$cookies,$rindex);
      
      $proxy = $1;
      $cookies = /cookies/io;
      if($proxy =~ /default/io){
	$rindex = 1;
      } else {
	$proxy =~ /([0-9\.]+):([0-9]+)/io;
	push(@proxies,scalar sockaddr_in($2,inet_aton($1)));
	$rindex = $#proxies;
      }
      while($_ = shift @commands){
	chomp;
	if(/^$/io){
	  last;
	}
	$routetab{$_} = [$rindex, $cookies];
      }	
    } else {
      print STDERR "Unknown config line: $_\n";
    }
  }
}
# Read configuration file into memory, following includes
sub readfile($){
    my ($posn,$file,$line,$filehandle);
    $file = shift;
    die "recursive config file $file\n" if($files { $file });
    $files { $file } ++;
    open FILE,$file or die "Can't read $file: $!\n";
    print "Reading config file $file...\n";
    while($line = <FILE>){
	if($line =~ /^include\s+(\S+)$/io){
	    $posn = tell FILE;
	    readfile($1); # recursive read of include file
	    open FILE,$file;
	    seek FILE,$posn,0;
	} else {
	    push @commands,$line;
	}
    }
    delete $files { $file };

}

# Parse HTTP standard date string, assumed to be in GMT

sub parsedate(){
  my $timestamp;
  my $datestr = shift;

  if($datestr =~ /([a-z]*),\s+([0-9]+)\s+(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)\s+([0-9]+)\s+([0-9]+):([0-9]+):([0-9]+)/io){
    $timestamp = timegm($7,$6,$5,$2,$monum{$3},$4);
  }
  return $timestamp;
}
